// Grainsring 2.0
// An approximate implementation of the Wiard Noisering with extra functionality for outputting triggers
//
// based on the Auduino code by Peter Knight and Ginkosynthese, revised and finetuned by Janne G:son Berg to convert it to noisering like functionality
//
// inputs
// Analog in 2: Chance; probability for new value to be 0 or 1
// Analog in 1: Change; probability that a new value is shifted in. 
// Analog in 0: Before 11 a clock, nothing happens.. 11 a clock, rotation and output as an 8 bit value. 12 a clock, rotation and lowest bit output for triggers and such. Trigger length the same as the high part of the incoming clock.
// Analog in 3: Incoming clock
// output
// Digital 11: Audio/trigger out (PWM)

#include <avr/io.h>
#include <avr/interrupt.h>

boolean clockHigh;
boolean previousClockHigh;
boolean change;
boolean chance;
boolean nextValue;

uint16_t bitArray;
uint16_t trigger;
uint16_t functionMode;

// Map Analogue channels
#define INCOMING_CLOCK       (3)
#define CHANGE_PROB_CONTROL  (1)
#define CHANCE_PROB_CONTROL  (2)
#define MODE_CONTROL         (0)

// Changing these will also requires rewriting audioOn()
//    Output is on pin 11
#define PWM_PIN       11
#define PWM_VALUE     OCR2A
#define LED_PIN       13
#define LED_PORT      PORTB
#define LED_BIT       5
#define PWM_INTERRUPT TIMER2_OVF_vect

void audioOn() {
  // Set up PWM to 31.25kHz, phase accurate
  TCCR2A = _BV(COM2A1) | _BV(WGM20);
  TCCR2B = _BV(CS20);
  TIMSK2 = _BV(TOIE2);
}

void setup() 
{
  // Initialize to 1000000010000000
  bitArray = 0x8080;
  // Initialize clock variables
  clockHigh = false;
  previousClockHigh = false;
  functionMode = 1000;

  pinMode(PWM_PIN,OUTPUT);
  audioOn();
  pinMode(LED_PIN,OUTPUT);
  //Serial.begin(9600);
}

void loop() 
{
  // int trigVal = analogRead(INCOMING_CLOCK);
  trigger = analogRead(INCOMING_CLOCK);

  // Going from low to high, ie incoming clock pulse
  if ((trigger > 32) && !clockHigh)
  {
    clockHigh = true;
  }

  // Going from high to low, end of incoming clock pulse
  if ((trigger <= 32) && clockHigh)
  {
    clockHigh = false;
  }

  functionMode = analogRead(MODE_CONTROL) >> 9;

  // Probabilities based on knob values
  change = analogRead(CHANGE_PROB_CONTROL) > random(1024);
  chance = analogRead(CHANCE_PROB_CONTROL) > random(1024);
}

SIGNAL(PWM_INTERRUPT)
{
  uint8_t output;

  // Only enter bit shift routine when incoming clock is rising for the first time
  if (!previousClockHigh && clockHigh)
  {
    // Rising edge of clock!
    previousClockHigh = true;
    // Rotate the bit array
    nextValue = bitRead(bitArray,15);
    bitArray = bitArray << 1;
    if (change)
    {
      nextValue = chance;
    }
    bitWrite(bitArray, 0, nextValue);
  }

  if (previousClockHigh && !clockHigh)
  {
    // Falling edge of clock!
    previousClockHigh = false;
  }

  // Either output value or on/off (bit) for trigger functionality
  if (functionMode == 0)
  {
    // Normal output
    output = lowByte(bitArray);
  }
  else
  {
    // Lowest bit output as trigger, only when incoming clock is high
    output = clockHigh && nextValue ? 255 : 0;
  }

  LED_PORT = nextValue << LED_BIT;
  
  // Output to PWM (this is faster than using analogWrite)  
  PWM_VALUE = output;
}

